/*
array.h

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991-96, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
*/

/*
This file originally had been generated with the command

   notangle -R'"array.h"' array.w > array.h

but is maintained as .h in libmawk.

(Notangle is part of Norman Ramsey's noweb literate programming package
available from CTAN(ftp.shsu.edu)).
*/

#ifndef ARRAY_H
#define ARRAY_H 1

#define AY_NULL         0
#define AY_INT          1
#define AY_STR          2
#define AY_SPLIT        4

#define NO_CREATE  0
#define CREATE     1

#include "array_orig.h"

/* create a new, empty array, copying implementation from imp; if imp is NULL,
   use the default (original) implementation. */
ARRAY new_ARRAY(mawk_state_t *MAWK, const array_imp_t *imp);

/* concat multiple expressions separated by SUBSEP to be an array index */
CELL *mawk_array_cat(mawk_state_t *MAWK, CELL *sp, int cnt);

/* clear the common administrative part of the array, leaving imp alone;
   this must be used instead of memsetting the ARRAY */
void mawk_array_clear_common(mawk_state_t *MAWK, ARRAY A);


/* wrappers for accessing the implementations; a layer of macros is introduced
   here to guarantee future flexibility */
/* **** WARNING: these macros will ecaluate ARR_ at least twice!  **** */
#define mawk_array_find(MAWK_, ARR_, cell_, alloc_)  (ARR_)->imp.find((MAWK_), (ARR_), (cell_), (alloc_))
#define mawk_array_delete(MAWK_, ARR_, cell_)        (ARR_)->imp.delet((MAWK_), (ARR_), (cell_))
#define mawk_array_clear(MAWK_, ARR_)                (ARR_)->imp.clear((MAWK_), (ARR_))
#define mawk_array_loop_vector(MAWK_, ARR_, sizep_)  (ARR_)->imp.loop_vect((MAWK_), (ARR_), (sizep_))
#define mawk_array_load(MAWK_, ARR_, cnt_)           (ARR_)->imp.load((MAWK_), (ARR_), (cnt_))

#endif /* ARRAY_H */
